"""Oracle Migrations.

Reason:
Oracle is a particularly picky database which requires a couple of changes before
Rasa X can run with it. Changes we had to do:

- Databases like MySQL or Oracle require a certain length for `VARCHAR` types.
  This migration fixes the length for all `String` columns. This is already done
  in the initial migration because this would otherwise break the migrations for Oracle.
  However, all database systems should have the same types so this migration applies the
  changes to existing databases.
- Oracle does not support index names with a length > 32. Hence, we have to rename the
  chars to shorter names.
- Oracle does not support auto-incrementing keys. We have to use sequences to get this
  working in Oracle.

Revision ID: 4daabca814ee
Revises: 6fce9679db61

"""

from alembic import op
import sqlalchemy as sa

from rasax.community.database import (
    AnalyticsCache,
    ChatToken,
    Conversation,
    PlatformFeature,
    Project,
    Role,
    ConversationActionMetadata,
    ConversationEntityMetadata,
    ConversationEvent,
    ConversationIntentMetadata,
    ConversationMessageCorrection,
    ConversationPolicyMetadata,
    ConversationSession,
    ConversationStatistic,
    Domain,
    Environment,
    Intent,
    Model,
    Permission,
    UserGoal,
    ConversationActionStatistic,
    ConversationEntityStatistic,
    ConversationIntentStatistic,
    ConversationPolicyStatistic,
    DomainAction,
    DomainEntity,
    DomainIntent,
    DomainSlot,
    MessageLog,
    ModelTag,
    NluEvaluation,
    SingleUseToken,
    TemporaryIntentExample,
    UserGoalIntent,
    NluEvaluationPrediction,
    User,
    LocalPassword,
)
from rasax.community.database.admin import user_role_mapping
from rasax.community.database.data import (
    EntitySynonym,
    LookupTable,
    RegexFeature,
    Story,
    Template,
    TrainingData,
    TrainingDataEntity,
)

import rasax.community.database.schema_migrations.alembic.utils as migration_utils

# revision identifiers, used by Alembic.
revision = "4daabca814ee"
down_revision = "6fce9679db61"
branch_labels = None
depends_on = None

FIXED_SIZE_COLUMNS = "fixed"
VARIABLE_SIZED_COLUMNS = "variable"

migrations = {
    AnalyticsCache: {
        FIXED_SIZE_COLUMNS: ["cache_key"],
        VARIABLE_SIZED_COLUMNS: ["result"],
    },
    ChatToken: {FIXED_SIZE_COLUMNS: ["bot_name", "description", "token"]},
    Conversation: {
        FIXED_SIZE_COLUMNS: ["sender_id", "latest_input_channel"],
        VARIABLE_SIZED_COLUMNS: ["evaluation"],
    },
    LocalPassword: {FIXED_SIZE_COLUMNS: ["password"]},
    PlatformFeature: {FIXED_SIZE_COLUMNS: ["feature_name"]},
    Project: {
        FIXED_SIZE_COLUMNS: ["project_id", "team", "active_model", "handoff_url"],
        VARIABLE_SIZED_COLUMNS: ["config"],
    },
    Role: {FIXED_SIZE_COLUMNS: ["role", "description"]},
    ConversationActionMetadata: {FIXED_SIZE_COLUMNS: ["conversation_id", "action"]},
    ConversationEntityMetadata: {FIXED_SIZE_COLUMNS: ["conversation_id", "entity"]},
    ConversationEvent: {
        FIXED_SIZE_COLUMNS: [
            "conversation_id",
            "type_name",
            "intent_name",
            "action_name",
            "policy",
        ],
        VARIABLE_SIZED_COLUMNS: ["data", "evaluation"],
    },
    ConversationIntentMetadata: {FIXED_SIZE_COLUMNS: ["conversation_id", "intent"]},
    ConversationMessageCorrection: {FIXED_SIZE_COLUMNS: ["conversation_id", "intent"]},
    ConversationPolicyMetadata: {FIXED_SIZE_COLUMNS: ["conversation_id", "policy"]},
    ConversationSession: {FIXED_SIZE_COLUMNS: ["conversation_id"]},
    ConversationStatistic: {FIXED_SIZE_COLUMNS: ["project_id"]},
    Domain: {FIXED_SIZE_COLUMNS: ["project_id", "path"]},
    EntitySynonym: {FIXED_SIZE_COLUMNS: ["name", "synonym", "project_id", "filename"]},
    Environment: {FIXED_SIZE_COLUMNS: ["name", "project", "url", "token"]},
    Intent: {FIXED_SIZE_COLUMNS: ["name", "mapped_to", "project_id"]},
    LookupTable: {
        FIXED_SIZE_COLUMNS: ["name", "filename", "project_id"],
        VARIABLE_SIZED_COLUMNS: ["elements"],
    },
    Model: {FIXED_SIZE_COLUMNS: ["hash", "name", "path", "project_id", "version"]},
    Permission: {FIXED_SIZE_COLUMNS: ["project", "role_id", "permission"]},
    User: {
        FIXED_SIZE_COLUMNS: [
            "username",
            "name_id",
            "team",
            "project",
            "authentication_mechanism",
            "attributes",
            {"name": "api_token", "size": 4000},
        ],
        VARIABLE_SIZED_COLUMNS: ["password_hash"],
    },
    RegexFeature: {FIXED_SIZE_COLUMNS: ["name", "filename", "project_id", "pattern"]},
    UserGoal: {FIXED_SIZE_COLUMNS: ["name", "project_id"]},
    ConversationActionStatistic: {FIXED_SIZE_COLUMNS: ["action", "project_id"]},
    ConversationEntityStatistic: {FIXED_SIZE_COLUMNS: ["entity", "project_id"]},
    ConversationIntentStatistic: {FIXED_SIZE_COLUMNS: ["intent", "project_id"]},
    ConversationPolicyStatistic: {FIXED_SIZE_COLUMNS: ["policy", "project_id"]},
    DomainAction: {FIXED_SIZE_COLUMNS: ["action"]},
    DomainEntity: {FIXED_SIZE_COLUMNS: ["entity"]},
    DomainIntent: {
        FIXED_SIZE_COLUMNS: ["intent", "triggered_action"],
        VARIABLE_SIZED_COLUMNS: ["ignore_entities", "use_entities"],
    },
    DomainSlot: {FIXED_SIZE_COLUMNS: ["slot", "initial_value", "type", "values"]},
    MessageLog: {
        FIXED_SIZE_COLUMNS: ["hash", "intent"],
        VARIABLE_SIZED_COLUMNS: ["intent_ranking", "entities", "text"],
    },
    ModelTag: {FIXED_SIZE_COLUMNS: ["tag"]},
    NluEvaluation: {
        FIXED_SIZE_COLUMNS: ["model_id"],
        VARIABLE_SIZED_COLUMNS: ["report"],
    },
    TrainingData: {
        FIXED_SIZE_COLUMNS: [
            "hash",
            "intent",
            "annotator_id",
            "project_id",
            "filename",
        ],
        VARIABLE_SIZED_COLUMNS: ["text"],
    },
    SingleUseToken: {FIXED_SIZE_COLUMNS: [{"name": "token", "size": 4000}, "username"]},
    Story: {
        FIXED_SIZE_COLUMNS: ["name", "user", "filename"],
        VARIABLE_SIZED_COLUMNS: ["story"],
    },
    Template: {
        FIXED_SIZE_COLUMNS: ["template", "text", "annotator_id", "project_id"],
        VARIABLE_SIZED_COLUMNS: ["content"],
    },
    TemporaryIntentExample: {FIXED_SIZE_COLUMNS: ["example_hash"]},
    UserGoalIntent: {FIXED_SIZE_COLUMNS: ["intent_name"]},
    user_role_mapping.name: {FIXED_SIZE_COLUMNS: ["username", "role"]},
    NluEvaluationPrediction: {FIXED_SIZE_COLUMNS: ["text", "intent", "predicted"]},
    TrainingDataEntity: {FIXED_SIZE_COLUMNS: ["entity", "value", "extractor"]},
}


def upgrade():
    # Migrate string columns to Oracle compatible types
    for table, new_types in migrations.items():
        table_name = table if isinstance(table, str) else table.__tablename__

        # Migrate columns which should get a fixed sized string type
        with op.batch_alter_table(table_name) as batch_op:
            for column in new_types.get(FIXED_SIZE_COLUMNS, []):
                if isinstance(column, str):
                    column_dict = {"name": column}
                else:
                    column_dict = column

                batch_op.alter_column(
                    column_dict["name"],
                    type_=sa.String(column_dict.get("size", 255)),
                    existing_type=sa.String(),
                )

        # Migrate columns which should get a variable sized string type
        if not isinstance(table, str):
            modifications = [
                migration_utils.ColumnTransformation(c, [sa.Text()], {"nullable": True})
                for c in new_types.get(VARIABLE_SIZED_COLUMNS, [])
            ]
            migration_utils.modify_columns(table_name, modifications)

    # Rename indexes which have been too long
    migration_utils.rename_or_create_index(
        "conversation_event",
        "conversation_event_conversation_id_index",
        "conv_event_conv_id_idx",
        ["conversation_id"],
    )
    migration_utils.rename_or_create_index(
        "conversation_event",
        "conversation_event_timestamp_index",
        "conv_event_timestamp_idx",
        ["timestamp"],
    )

    # Rename wrong index name
    from rasax.community.database.schema_migrations.alembic.versions.migration_2019_07_05_indexes_for_logs_and_nlu_training_data_b4a9bd47b6e0 import (
        NLU_TRAINING_DATA_INDEX_NAME,
    )

    migration_utils.rename_or_create_index(
        "nlu_training_data",
        "NLU_TRAINING_DATA_INDEX_NAME",
        NLU_TRAINING_DATA_INDEX_NAME,
        ["hash"],
    )

    # Fix wrong index
    from rasax.community.database.schema_migrations.alembic.versions.migration_2019_07_22_add_index_to_event_timestamps_45546b047abe import (
        NEW_INDEX_NAME,
    )

    migration_utils.rename_or_create_index(
        "conversation_event", NEW_INDEX_NAME, NEW_INDEX_NAME, ["timestamp"]
    )

    # Create `Sequence`s which provide the auto-incrementing id for Oracle
    tables_with_sequences = [
        "conversation_event",
        "message_log",
        "template",
        "story",
        "nlu_training_data",
        "nlu_training_data_entity",
        "regex_feature",
        "lookup_table",
        "domain",
        "domain_action",
        "domain_intent",
        "domain_entity",
        "domain_slot",
        "intent",
        "temporary_intent_example",
        "user_goal",
        "model",
        "nlu_evaluation",
        "nlu_evaluation_prediction",
    ]
    for table in tables_with_sequences:
        migration_utils.create_sequence(table)


def downgrade():
    for table, new_types in migrations.items():
        table_name = table if isinstance(table, str) else table.__tablename__
        with op.batch_alter_table(table_name) as batch_op:
            for column in new_types.get(FIXED_SIZE_COLUMNS, []):
                batch_op.alter_column(
                    column, type_=sa.String(), existing_type=sa.String(255)
                )

        if not isinstance(table, str):
            modifications = [
                migration_utils.ColumnTransformation(
                    c, [sa.String()], {"nullable": True}
                )
                for c in new_types.get(VARIABLE_SIZED_COLUMNS, [])
            ]
            migration_utils.modify_columns(table_name, modifications)

    migration_utils.rename_or_create_index(
        "conversation_event",
        "conv_event_conv_id_idx",
        "conversation_event_conversation_id_index",
        ["conversation_id"],
    )
    migration_utils.rename_or_create_index(
        "conversation_event",
        "conv_event_timestamp_idx",
        "conversation_event_timestamp_index",
        ["conversation_id"],
    )
