import os

import pytest
from jwt import InvalidSignatureError

import rasax.community.database.utils as db_utils
import rasax.community.jwt
from rasax.community import config, constants
from rasax.community import utils
from tests.unit.conftest import generate_key_pair


def test_verify_bearer_token(tmpdir_factory):
    # get valid token
    private_key_path, public_key_path = generate_key_pair(tmpdir_factory, 0)
    private_key = utils.file_as_bytes(private_key_path)
    public_key = utils.file_as_bytes(public_key_path)

    payload = {"test": "test"}
    encoded = rasax.community.jwt.bearer_token(payload, private_key=private_key)

    assert (
        rasax.community.jwt.verify_bearer_token(encoded, public_key=public_key)
        == payload
    )


def test_verify_invalid_bearer_token():
    # No authorisation header
    with pytest.raises(TypeError):
        rasax.community.jwt.verify_bearer_token(None)

    # authorization header does not contain 'Bearer ' prefix
    header_value = "bearer SOME_TOKEN"
    with pytest.raises(ValueError):
        rasax.community.jwt.verify_bearer_token(header_value)


def test_verify_bearer_token_invalid_public_key(tmpdir_factory):
    private_key_path, public_key_path = generate_key_pair(tmpdir_factory, 0)
    private_key = utils.file_as_bytes(private_key_path)

    payload = {"test": "test"}
    encoded = rasax.community.jwt.bearer_token(payload, private_key=private_key)

    with pytest.raises(ValueError) as exception_info:
        rasax.community.jwt.verify_bearer_token(encoded, public_key="WRONG_KEY")

    # make sure the last `ValueError` in `verify_bearer_token()` was raised
    assert "Invalid bearer token" in str(exception_info.value)


def test_invalid_bearer_tokens():
    with pytest.raises(ValueError):
        rasax.community.jwt.bearer_token({}, private_key="wrong secret")

    with pytest.raises(ValueError):
        rasax.community.jwt.encode_jwt({}, private_key="some random key")

    with pytest.raises(ValueError):
        rasax.community.jwt.verify_bearer_token("", public_key="wrong key")


# noinspection PyProtectedMember
def test_jwt_keys_from_file(tmpdir_factory):
    # generate rsa keypair
    private_key_path, public_key_path = generate_key_pair(tmpdir_factory, 0)

    # verification should work
    private_key, public_key = rasax.community.jwt._fetch_and_verify_jwt_keys_from_file(
        private_key_path, public_key_path
    )
    rasax.community.jwt._set_keys(private_key, public_key)

    # check that environment variables are set
    assert config.jwt_private_key == utils.file_as_bytes(private_key_path)
    assert config.jwt_public_key == utils.file_as_bytes(public_key_path)

    # wrong key path
    with pytest.raises(SystemExit):
        _ = rasax.community.jwt._fetch_and_verify_jwt_keys_from_file(
            private_key_path, "some/wrong/path"
        )

    # remove -----BEGIN PRIVATE KEY----- header from private key
    tag = "-----BEGIN PRIVATE KEY-----"
    assert tag.encode() in utils.file_as_bytes(private_key_path)
    os.system("sed -i.bak '/{}/d' {}".format(tag, private_key_path))
    assert tag.encode() not in utils.file_as_bytes(private_key_path)

    private_key = utils.file_as_bytes(private_key_path)
    public_key = utils.file_as_bytes(public_key_path)
    with pytest.raises(ValueError):
        rasax.community.jwt._verify_keys(private_key, public_key)

    # generate new private and public keys which don't match each other
    new_private_key_path, _ = generate_key_pair(tmpdir_factory, 1)
    _, new_public_key_path = generate_key_pair(tmpdir_factory, 2)

    new_public_key = utils.file_as_bytes(new_public_key_path)
    new_private_key = utils.file_as_bytes(new_private_key_path)
    with pytest.raises(InvalidSignatureError):
        rasax.community.jwt._verify_keys(new_private_key, new_public_key)


# noinspection PyProtectedMember
def test_generate_key_pair():
    private_key, public_key = rasax.community.jwt._generate_rsa_key_pair(2048)
    rasax.community.jwt._verify_keys(private_key, public_key)


# noinspection PyProtectedMember
def test_generate_and_set_jwt_keys():
    private_key, public_key = rasax.community.jwt._generate_rsa_key_pair()
    rasax.community.jwt._set_keys(private_key, public_key)
    assert config.jwt_public_key == public_key
    assert config.jwt_private_key == private_key


# noinspection PyProtectedMember
def test_save_private_key_to_temp_path():
    private_key, _ = rasax.community.jwt._generate_rsa_key_pair()
    path = rasax.community.jwt._save_rsa_private_key_to_temporary_file(private_key)
    with open(path, "rb") as f:
        assert f.read() == private_key


# noinspection PyProtectedMember
def test_add_jwt_key_to_dictionary():
    private_key, public_key = rasax.community.jwt._generate_rsa_key_pair()
    rasax.community.jwt._set_keys(private_key, public_key)

    d = {}
    rasax.community.jwt.add_jwt_key_to_result(d)
    assert d["keys"]

    jwt = d["keys"][0]
    assert all(k in jwt for k in ["alg", "key"])

    assert jwt["key"] == config.jwt_public_key
    assert jwt["alg"] == constants.JWT_METHOD
